/*
 * advanced tacacs+ server based on Cisco Systems tac_plus code.
 * For Cisco copyright notice, see LICENSE.CISCO
 *
 * This is a copyrighted commercial software; see LICENSE for details.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <arpa/nameser.h>
#include <setjmp.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#ifdef HAVE_DB_185_H
#include <db_185.h>
#else
#include <db.h>
#endif
#include "dynipd.h"
#include "allow.h"

extern int sockread(int, u_char *, int, int);

#ifndef O_SHLOCK
#ifdef O_NSHARE
#define O_SHLOCK O_NSHARE
#else
#define O_SHLOCK 0
#endif
#endif

#define NHASH           128
#define NAPOOLS         64

struct htab {                           /* two-way hashtable      */
	struct  ud      *udp;           /* pointer to user struct */
	struct  htab    *next;          /* chain link             */
	int             active;         /* active flag            */
}
	*htabi[NHASH],                  /* references by ipns     */
	*htabn[NHASH];                  /* references by names    */

#define H_GET           0001            /* lookup                 */
#define H_UPDATE        0002            /* create/update          */
#define H_DELETE        0004            /* delete entry           */

/* Common structure for in-core user data */
struct  ud {
	time_t          time;           /* allocation time    */
	u_long          addr;           /* assigned address   */
	char           *name;           /* User name          */
	char           *pname;          /* pool name ref      */
};

struct udb_data {                       /* Data being held in external db */
	time_t          time;           /* allocation time                */
	u_long          addr;           /* assigned address               */
};

struct dqueue {                         /* Database update stack          */
	struct  ud     *udp;            /* pointer to user structure      */
	int             op;             /* Operation                      */
	struct dqueue  *next;
	struct dqueue  *prev;
};

static struct dqueue *dq_head = (struct dqueue *) 0,
		     *dq_tail = (struct dqueue *) 0;

struct apool {
	u_long          a1;             /* Beginning of range */
	u_long          a2;             /* End of range       */
	int             ps;             /* Range size         */
	u_long         *stack;          /* freelist stack     */
	int             sp;             /* stack ptr          */
	struct plist   *pool;           /* pool backref       */
};

struct plist {                          /* Named pools list    */
	int             curpool;        /* for circular search */
	char           *pname;          /* pool list name      */
	int             nranges;        /* number of pools     */
	struct apool    apools[NAPOOLS];/* apool list          */
	struct plist   *next;           /* ptr to next element */
};
static struct plist *p_list = (struct plist *) 0;

/* some defaults */
#ifndef CONFPATH
#define CONFPATH "/etc"
#endif
#ifndef DATAPATH
#define DATAPATH "/var/adm/tac_plus"
#endif
#ifndef LOGPATH
#define LOGPATH DATAPATH
#endif

#ifndef CDBNAME
#define CDBNAME DATAPATH "/ipaddr.db"
#endif
#ifndef CCONFNAME
#define CCONFNAME CONFPATH "/dynipd.conf"
#endif
#ifndef DYNLOGFILE
#define DYNLOGFILE LOGPATH "/dynipd.log"
#endif

#define PTMAX           5               /* secs - protocol timeout */

#define NTOA(s,n)       (void) sprintf((s), "%ld.%ld.%ld.%ld",  \
	((n) >> 24) & 0xFF, ((n) >> 16) & 0xFF,                 \
	((n) >> 8 ) & 0xFF, (n) & 0xFF)

#define SANE_RANGE      16384           /* how many addresses ...    */

static DB       *cachedb = NULL;        /* cache DB handle         */
static char     cachedbname[512] = "";  /* cache DB name           */
static int      debug   = 0;            /* debugging               */
static int      inter   = 0;            /* non-background          */
static int      udb_update = 15;        /* dbupdate - 30 secs dflt */
static char     confname[FILENAME_MAX] = CCONFNAME;
static u_long   Seq;                    /* Sequence #              */
static int      dynipd_port = 0;        /* daemon port number      */
static int      proto_clock = -1,
		upd_clock   = 0;
static int      my_pid;
sigjmp_buf      abrt;
static char     Logfile[512] = DYNLOGFILE;
static char     default_pool_name[] = "default";

void    parse_cf(void), report(int, char *, ...), doexit(int);
int     dyna_setrange(char *, char *);
struct  ud *hashi(int, u_long, struct ud *),
	*hashn(int, char *, struct ud *);
void    *lmalloc(int);
u_long  dyna_alloc(char *, struct plist *), dyna_lalloc(struct plist *);
void    udb_sync(), udb_load(), dyna_invalidate(struct plist *), myclock();
void    udb_queue_request(int, struct ud *);
struct  apool *getap(u_long);

/*
 * Options:
 *      -g              - debug mode; don't go to background
 *      -I              - interactive mode; don't go to background
 *      -c <confname>   - alternate location of config. file
 *      -D <dbname>     - altername cachedb location, overrides cf
 *      -L <logfile>    - alternate log location
 *      -p port         - listen on alternate port
 */

int main(argc, argv)
	int     argc;
	char    *argv[];
{
	int                     i, sc, sd, one = 1;
	size_t	psalen;
	struct  sockaddr_in     isa, psa;
	struct  itimerval       it[1];
	struct  apool           *ap;
	struct  plist           *pp;
	u_long                  l;
	struct dynp             dpb;
	char                    username[MAXARGBUF/2],
				poolname[MAXARGBUF/2];

	while((i = getopt(argc, argv, "gIc:D:p:L:")) != -1) {
		switch (i) {
			case 'g':
				debug = 1;
				continue;
			case 'I':
				inter = 1;
				continue;
			case 'c':
				(void) strcpy(confname, optarg);
				continue;
			case 'D':
				(void) strcpy(cachedbname, optarg);
				continue;
			case 'L':
				(void) strcpy(Logfile, optarg);
				continue;
			case 'p':
				dynipd_port = atoi(optarg);
				continue;
			case '?':
usage:
				report(2, "Usage: %s [-g] [-I] [-c conf] [-D cachedb]\n",
					argv[0]);
		}
	}
	my_pid = getpid();
	if (optind < argc)
		goto    usage;

	parse_cf();                     /* read config */

	/* Initialize stacks */
	for (pp = p_list; pp; pp = pp->next) {
		for (ap = pp->apools; ap < &pp->apools[pp->nranges]; ap++) {
			ap->stack = (u_long *)
			    lmalloc((ap->ps + 1) * sizeof (u_long));
			ap->sp    = -1;
		}
	}
	/* Daemonize */
	if (!debug && !inter) {
		if (fork())
			exit(0);
		(void) setsid();
	}

	if (!dynipd_port)
		dynipd_port = DPORT;

	(void) signal(SIGTERM, doexit);
	(void) signal(SIGINT,  doexit);
	(void) signal(SIGALRM, myclock);
	(void) signal(SIGCHLD, SIG_IGN);

	/* set timer */
	it->it_value.tv_sec  = 1;
	it->it_value.tv_usec = 1;
	it->it_interval = it->it_value;

	(void) setitimer(ITIMER_REAL, it, NULL);         /* launch clock */

	udb_load();                     /* load current map */

#ifdef DEBUG
	testloop();
#endif
	/* Do all socket stuff */
	if ((sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
		report(2, "can't get socket");

	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
	    (char *) &one, sizeof (one)) < 0)
		report(2, "setsockopt REUSEADDR: %s\n", strerror(errno));

	if (setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE,
	    (char *) &one, sizeof (one)) < 0)
		report(2, "setsockopt KEEPALIVE: %s\n", strerror(errno));

	bzero((char *) &isa, sizeof (isa));
	isa.sin_family = AF_INET;
	isa.sin_addr.s_addr = htonl(INADDR_ANY);
	isa.sin_port = htons(dynipd_port);

	if (bind(sd, (struct sockaddr *) &isa, sizeof (isa)) < 0)
		report(2, "can't bind to port %d: %s\n",
			dynipd_port, strerror(errno));

	if (listen(sd, 10) < 0)
		report(2, "listen returned error: %s\n",strerror(errno));

	report(1, "Daemon mode; listening on port %d\n", dynipd_port);

	sc = -1;
	for (;;) {
		(void) sigsetjmp(abrt, 1);
		if (sc > 0)
			(void) close(sc);
		psalen = sizeof(psa);
		if ((sc = accept(sd, (struct sockaddr *) &psa, &psalen)) < 0) {
#ifndef BSD
			if (errno==EINTR) continue;
#endif
			report(1, "accept() failed, %s\n", strerror(errno));
			continue;
		}
		if(allowaddr(psa.sin_addr.s_addr))
			report(0, "accepted call from %s\n",
				inet_ntoa(psa.sin_addr));
		else {
			report(1, "Unauthorized call from %s\n",
				inet_ntoa(psa.sin_addr));
			continue;
		}
		proto_clock = 0;
		(void) read(sc, (char *) &dpb, 1);
		if (sockread(sc, (u_char *) &dpb, sizeof (dpb), 5) == -1) {
			report(1, "protocol error: %s\n", strerror(errno));
			continue;
		}

		if (dpb.magic != DYNP_MAGIC ||
		    (dpb.op != DYNP_IARQ && dpb.op != DYNP_IARQ2)) {
			report(1, "bad magic in incoming packet\n");
			continue;
		}
		(void) write(sc, &one, 1);      /* ACK */
		proto_clock = -1;

		strncpy(username, dpb.argb, (MAXARGBUF/2) - 1);
		if (dpb.op == DYNP_IARQ2)
			strcpy(poolname, &dpb.argb[MAXARGBUF/2]);
		else
			strcpy(poolname, default_pool_name);
		for (pp = p_list; pp; pp = pp->next)
			if (!strcmp(pp->pname, poolname))
				break;
		if (pp) {
			l = dyna_alloc(username, pp);
			if (l == 0) {
			    (void) sprintf(dpb.argb, "no db entry found for for %s [static]\n", username);
			    dpb.op = DYNP_IERROR;
			} else {
			    NTOA(dpb.argb, l);
			    dpb.op = DYNP_IAREPL;
			}
		} else {
			(void) sprintf(dpb.argb, "Pool %s not found\n", poolname);
			dpb.op = DYNP_IERROR;
			report(1, dpb.argb);
		}
		proto_clock = 0;
		if (write(sc, (char *) &dpb, sizeof (dpb)) != sizeof (dpb)) {
			report(1, "error replying to request\n");
			continue;
			/* we should better deallocate address here ... */
		}
		if (dpb.op == DYNP_IAREPL)
			report(1, "alloc: %s <- %s [%s]\n", dpb.argb,
				username, poolname);
		proto_clock = -1;
	}
}

/*
 * Allocate address for particular user
 */
u_long  dyna_alloc(user, pool)
char           *user;
struct plist   *pool;
{
	struct ud       *udp, uda[1];
	struct apool    *ap;

	/* check in hashtab first */
	while((udp = hashn(H_GET, user, NULL))) {
		if (udp->pname == pool->pname) {
			udp->time = Seq++;                  /* refresh */
			udb_queue_request(H_UPDATE, udp);
			return (udp->addr);
		}
		if (!(ap = getap(udp->addr)))
			report(2, "internal err: addr %ld not in any range\n",
			    udp->addr);
		ap->stack[++ap->sp] = udp->addr;
		udb_queue_request(H_DELETE, udp);
		(void) hashn(H_DELETE, user, NULL);
		report(1, "info: pool changed %s -> %s for user %s\n", udp->pname, pool->pname, user);
	}
	if (!strncmp(pool->pname,"static",6)) {
	    report(1, "error: no db entry found for for %s [static]\n", user);
	    return 0;
	}
	/* not in cache -- allocate locally and create new entry */
	uda->name  = (char *) lmalloc(strlen(user) + 1);
	uda->time  = Seq++;
	uda->addr  = dyna_lalloc(pool);
	uda->pname = pool->pname;
	(void) strcpy(uda->name, user);

	udb_queue_request(H_UPDATE, uda);
	(void) hashn(H_UPDATE, user, uda);

	return (uda->addr);
}

/*
 * Get next number from local stack
 */
u_long  dyna_lalloc(pool)
struct plist    *pool;
{
	struct  apool   *ap;
	int     oldc = pool->curpool;

	for (;;) {
		for (ap = &pool->apools[pool->curpool];
		    pool->curpool < pool->nranges; ap++, pool->curpool++)
			if (ap->sp >= 0)
				return (ap->stack[ap->sp--]);
		pool->curpool = 0;
		for (ap = pool->apools;
		    pool->curpool < oldc; ap++, pool->curpool++)
			if (ap->sp >= 0)
				return (ap->stack[ap->sp--]);

		dyna_invalidate(pool); /* _should_ add at least one element */
	}
}

/*
 * Read and parse configuration file.
 * Configuration file syntax:
 *
 * db     <dbfilename>            - define db for cache
 * range  x.x.x.x y.y.y.y         - define pool range; multiple allowed
 * update <seconds>               - db update interval in secs
 * allow-from <ipaddr>            - accept connections only from given host
 */
void    parse_cf()
{
	FILE    *cfp;
	char    *p, *q, iline[128];
	int     lnum = 0;

	if (!(cfp = fopen(confname, "r")))
		report(2, "can't open config file %s\n", confname);
	while(fgets(iline, 128, cfp)) {
		lnum++;
		for (p = iline; *p != '\n' && *p != '#' && *p; p++)
			;
		*p = '\0';
		if (!*iline)
			continue;
		for (p = iline; *p != ' ' && *p != '\t' && *p; p++)
			;
		if (!*p)
synt:
			report(2, "syntax error in config file line %d\n", lnum);
		*p = '\0';
		for (q = ++p; *q == ' ' || *q == '\t'; q++);
			;
		if (!*q)
			goto synt;
		if (!strcmp(iline, "db")) {
			(void) strcpy(cachedbname, q);
		} else if (!strcmp(iline, "update")) {
			if (sscanf(q, "%d", &udb_update) != 1)
				goto synt;
			if (udb_update < 5)
				report(2, "too small update period\n");
		} else if (!strcmp(iline, "range")) {
			if (dyna_setrange(default_pool_name, q))
				goto synt;
		} else if (!strcmp(iline, "pool")) {
			for (p = q; *p != ' ' && *p != '\t' && *p; p++)
				;
			if (!p[0] || !p[1])
				goto synt;
			*p++ = '\0';
			while(*p == ' ' || *p == '\t')
				p++;
			if (dyna_setrange(q, p))
				goto synt;
		} else if (!strcmp(iline, "logfile")) {
			if (!strcmp(Logfile, DYNLOGFILE))
				(void) strcpy(Logfile, q);
		} else if (!strcmp(iline, "allow-from")) {
			if (put_new_allow_addr(q)) {
				report(1, "bad ip address %s\n", q);
				goto synt;
			}
		} else if (!strcmp(iline, "port")) {
			int     port = atoi(q);

			if (port <= 0 || port > 65535) {
				report(1, "bad port %d\n", port);
				goto synt;
			}
			if (!dynipd_port)
				dynipd_port = port;
		}
	}
	(void) fclose(cfp);
}

/*
 * Load external db into local hash
 */
void    udb_load()
{
	struct  apool   *ap;
	struct  plist   *pp;
	DBT             k[1], d[1];
	struct  ud      udp[1];
	u_long          l;
	int             lfree = 0;

	cachedb = dbopen(cachedbname, O_RDWR, 0600, DB_HASH, NULL);
	if (!cachedb) {
		(void) unlink(cachedbname);
		cachedb = dbopen(cachedbname, O_CREAT|O_RDWR, 0600, DB_HASH, NULL);
		if (!cachedb)
			report(2, "cannot open db %s\n", cachedbname);
	}

	while(!(*cachedb->seq) (cachedb, k, d, R_NEXT)) {
		udp->name = (char *) lmalloc(k->size + 1);
		(void) strncpy(udp->name, k->data, k->size);
		udp->name[k->size]='\0';
		udp->addr = ((struct udb_data *) d->data)->addr;
		udp->time = ((struct udb_data *) d->data)->time;

		if (udp->time > Seq)
			Seq = udp->time;

		if (!(ap = getap(udp->addr))) {
			/* not from interval */
			(void) free(udp->name);
			(*cachedb->del) (cachedb, k, NULL);
			continue;
		}
		/* maybe we need to be more discrete here? */
		if (hashi(H_GET, udp->addr, NULL)) {
			report(1, "duplicate addr in db: %08X\n", udp->addr);
			(void) free(udp->name);
			(*cachedb->del) (cachedb, k, NULL);
		} else {
			udp->pname = ap->pool->pname; /* pool reference */
			(void) hashn(H_UPDATE, udp->name, udp);
		}
	}
	(*cachedb->sync) (cachedb, NULL);
	Seq++;

	/* Fill the stacks */
	for (pp = p_list; pp; pp = pp->next) {
		lfree = 0;
		for (ap = pp->apools; ap < &pp->apools[pp->nranges]; ap++) {
			ap->sp = -1;
			for (l = ap->a1; l <= ap->a2; l++)
				if (!hashi(H_GET, l, NULL)) {
					ap->stack[++ap->sp] = l;
					lfree++;
				}
		}
		if (lfree <= 0)
			dyna_invalidate(pp);              /* uh oh */
	}
}

/*
 * compare elements of udp pools
 */
int     dyna_comp(u1, u2)
	const void *u1, *u2;
{
	if (((struct ud*)(u1))->time > ((struct ud*)(u2))->time)
		return (1);
	if (((struct ud*)(u1))->time < ((struct ud*)(u2))->time)
		return (-1);
	return 0;
}

/*
 * Most complex part - cache invalidation algorithm.
 *      - create array of pointers to udb entries
 *      - sort this array by time
 *      - take some elements from the top of array and add it to pool stacks
 */
void    dyna_invalidate(pool)
struct plist    *pool;
{
	struct apool    *ap;
	struct ud       **upool, *udp;
	int             i, top;

	for (ap = pool->apools; ap < &pool->apools[pool->nranges]; ap++) {
		upool = (struct ud **)
			lmalloc(sizeof(struct ud *) * (ap->ps + 1));

		/* fill the pool */
		for (i = 0; i < ap->ps; i++)
			if (!(upool[i] = hashi(H_GET, ap->a1 + i, NULL)))
				report(2, "internal err - dynainv %d/%08X\n",
					i, ap->a1);
		qsort(upool, ap->ps, sizeof (struct ud *), dyna_comp);
		top = ap->ps / 3;
		if (top < 1)
			top = 1;
		ap->sp = -1;
		for (i = 0; i < top; i++) {
			udp = upool[i];
			ap->stack[++ap->sp] = udp->addr;
			udb_queue_request(H_DELETE, udp);
			(void) hashn(H_DELETE, udp->name, NULL);
		}
	}
	pool->curpool = 0;
}

/*
 * Hash-table reference by name
 */
struct  ud      *hashn(op, name, udp)
	int             op;
	char           *name;
	struct ud      *udp;

{
	int             k;
	unsigned int    hval = 0;
	struct htab    *hp;

	k = strlen(name);
	while (--k >= 0)
		hval += name[k] * (k + 1);
	hval %= NHASH;

	for (hp = htabn[hval]; hp; hp = hp->next)
		if (!strcmp(name, hp->udp->name))
		       break;
	switch (op) {

	    case H_GET:
		return (hp && hp->active ? hp->udp : (struct ud *) 0);

	    case H_DELETE:
		if (hp) {
			hp->active = 0;
			(void) hashi(H_DELETE, hp->udp->addr, NULL);
		}
		return ((struct ud *) 0);

	    case H_UPDATE:
		if (hp) {
			hp->active = 1;
			(void) hashi(H_DELETE, hp->udp->addr, NULL);
			hp->udp->time = udp->time;
			hp->udp->addr = udp->addr;
			(void) hashi(H_UPDATE, udp->addr, hp->udp);
			return (hp->udp);
		}
		hp = (struct htab *) lmalloc(sizeof (struct htab));
		hp->udp       = (struct ud *) lmalloc(sizeof (struct ud));
		hp->udp->name = udp->name;
		hp->udp->addr = udp->addr;
		hp->udp->time = udp->time;
		hp->udp->pname = udp->pname;
		hp->next      = htabn[hval];
		hp->active    = 1;
		htabn[hval]   = hp;
		(void) hashi(H_UPDATE, udp->addr, hp->udp);
		return (hp->udp);
	}
	return ((struct ud *) 0);
}

/*
 * Hash-table reference by IP number
 */
struct  ud      *hashi(op, inum, udp)
	int             op;
	u_long          inum;
	struct ud      *udp;
{
	struct htab    *hp;
	unsigned int    hval = ((inum * 1333 + 5) * 11) % NHASH;

	for (hp = htabi[hval]; hp; hp = hp->next)
		if (inum == hp->udp->addr)
			break;
	switch (op) {

	    case H_GET:
		return (hp ? hp->udp : (struct ud *)0);

	    case H_DELETE:
		if (hp)
			hp->active = 0;
		return ((struct ud *) 0);

	    case H_UPDATE:
		if (hp) {
			hp->active = 1;
			hp->udp = udp;
			return (udp);
		}
		hp = (struct htab *) lmalloc(sizeof (struct htab));
		hp->udp    = udp;
		hp->next   = htabi[hval];
		hp->active = 1;
		htabi[hval] = hp;

		return (udp);
	}
	return ((struct ud *) 0);
}

/*
 * get a pointer to corresponding address range
 */
struct apool *getap(addr)
	u_long          addr;
{
	struct plist    *pp;
	struct apool    *ap;

	for (pp = p_list; pp; pp = pp->next) {
		for (ap = pp->apools; ap < &pp->apools[pp->nranges]; ap++) {
			if (addr >= ap->a1 && addr <= ap->a2)
				return ap;
		}
	}
	return (struct apool *) 0;
}

#define VALID_IPN(n)    ((n) >= 0 && (n) <= 255)

int     dyna_setrange(pname, s)
	char    *pname, *s;
{
	struct  apool   *ap;
	struct  plist   *pp;
	int             i = 0, ib[8];

	/* check for pool "pname" */
	for (pp = p_list; pp; pp = pp->next)
		if (!strcmp(pp->pname, pname))
			break;
	if (!pp) {
		pp = (struct plist *) lmalloc(sizeof(struct plist));
		pp->pname = (char *) lmalloc(strlen(pname) + 1);
		(void) strcpy(pp->pname, pname);
		pp->curpool = 0;
		pp->nranges = 0;
		pp->next = p_list;
		p_list = pp;
	}
	ap = &pp->apools[pp->nranges];
	ap->pool = pp;
	if (sscanf(s, "%d.%d.%d.%d %d.%d.%d.%d", &ib[3], &ib[2],
	    &ib[1], &ib[0], &ib[7], &ib[6], &ib[5], &ib[4]) != 8)
		goto synt1;

	for (i = 0; i < 4; i++) {
		if (!VALID_IPN(ib[i]) || !VALID_IPN(ib[i + 4])) {
synt1:
			report(1, "bad ip number(s) in cf file\n");
			return (1);
		}
		ap->a1 |= (ib[i]     << (i << 3));
		ap->a2 |= (ib[i + 4] << (i << 3));
	}
	ap->ps = ap->a2 - ap->a1;
	if (ap->ps < 0 || ap->ps > SANE_RANGE) {
		report(1, "bad IPN range in cf file\n");
		return (1);
	}
	ap->ps++;
	if (getap(ap->a1) || getap(ap->a2))
		report(2, "error: address ranges overlap!\n");
	if (++pp->nranges >= NAPOOLS)
		report(2, "error: too many ranges defined for pool %s\n",
		    pp->pname);
	return (0);
}

/*
 * database update stuff
 */
void    udb_queue_request(op, udp)
	int             op;
	struct ud       *udp;
{
	struct  dqueue  *dq;

	/* add request to dq_list chain link */
	dq = dq_head;
	dq_head = (struct dqueue *) lmalloc(sizeof(struct dqueue));
	if (!dq)
		dq_tail = dq_head;
	dq_head->next = dq;
	dq_head->prev = (struct dqueue *) 0;
	if (dq)
		dq->prev = dq_head;

	dq_head->udp  = (struct ud *) lmalloc(sizeof (struct ud));
	*(dq_head->udp) = *udp;
	dq_head->udp->name = (char *) lmalloc(strlen(udp->name) + 1);
	(void) strcpy(dq_head->udp->name, udp->name);
	dq_head->op   = op;
}

/*
 * Actually update DB
 */
void    udb_doupdate()
{
	struct  dqueue   *dq, *dp;
	struct  udb_data udd[1];
	DBT     k[1], d[1];
	char    tkey[128];

	report(0, "udb_doupdate called\n");

	if (!dq_tail)
		return;

	for (dq = dq_tail; dq; ) {
		k->size = strlen(dq->udp->name);
		k->data = tkey;
		(void) strcpy(k->data, dq->udp->name);

		if (dq->op == H_DELETE) {
			(*cachedb->del) (cachedb, k, NULL);
		} else if (dq->op == H_UPDATE) {
			udd->time = dq->udp->time;
			udd->addr = dq->udp->addr;
			d->data   = udd;
			d->size   = sizeof (struct udb_data);
			if ((*cachedb->put) (cachedb, k, d, NULL) == -1)
				report(1, "can't update db!\n");
		}
		dp = dq;
		dq = dq->prev;
		(void) free(dp->udp->name);
		(void) free(dp->udp);
		(void) free(dp);
	}
	dq_head = dq_tail = (struct dqueue *) 0;
	(*cachedb->sync) (cachedb, NULL);
}

/*
 * Local time-related events handler
 */
void    myclock()
{
	(void) signal(SIGALRM, SIG_IGN);

	if (upd_clock++ >= udb_update) {
		upd_clock = 0;
		udb_doupdate();
	}
	if (proto_clock >= 0 && proto_clock++ >= PTMAX) {
		proto_clock = -1;
		report(1, "Session aborted - timeout!\n");
		(void) signal(SIGALRM, myclock);
		siglongjmp(abrt, 0);
	}
	(void) signal(SIGALRM, myclock);
}

/*
 * Logging stuff
 */
void    report(int lvl, char *fmt,...)
{
	va_list         ap;
	char            msg[1024];      /* longer than any other buffers */
	int             logfd;

	va_start(ap, fmt);
	(void) vsprintf(msg, fmt, ap);
	va_end(ap);

	if (debug)
		fputs(msg, stderr);

	if (debug || lvl > 0) {
	    logfd = open(Logfile, O_CREAT | O_WRONLY |
		O_APPEND | O_SHLOCK, 0666);
	    if (logfd >= 0) {
		char buf[1024];
		time_t t;

		(void) time(&t);
		(void) strcpy(buf, (char *) ctime(&t));
		buf[24] = ' ';
		(void) sprintf(&buf[25], "[%d]: %s", my_pid, msg);

		write(logfd, buf, strlen(buf));
		close(logfd);
	    }
	}
	if (lvl > 1)
		doexit(lvl);
}

/* memory allocator */
void    *lmalloc(size)
	int     size;
{
	void    *p = (void *) malloc(size);

	if (!p)
		report(2, "can't allocate memory, exiting\n");
	return (p);
}

/*
 * doexit -- clean exit
 */
void    doexit(n)
	int     n;
{
	if (cachedb) {
		udb_doupdate();
		(*cachedb->close) (cachedb);
	}
	exit(n);
}


#ifdef DEBUG

testloop()
{
	char    *p, *q, ibuf[128], tmp[128];
	struct  htab     *hp;
	struct  plist    *pp;
	u_long  l;

	for (;;) {
#if 0
		printf(" --- hash tables ---\nnhash:");
		for (l = 0; l < NHASH; l++) {
			for (hp = htabn[l]; hp; hp = hp->next)
				printf("-%d<%s> %08X %d-", hp->active, hp->udp->name,
					hp->udp->addr, hp->udp->time);
		}
		printf("\nihash:\n");
		for (l = 0; l < NHASH; l++) {
			for (hp = htabi[l]; hp; hp = hp->next)
				printf("-%d<%s> %08X %d-", hp->active, hp->udp->name,
					hp->udp->addr, hp->udp->time);
		}
		printf("\n------------------------------\n\n");
#endif

		fputs("dynipd> ", stdout);

		if (!fgets(ibuf, 128, stdin))
			doexit(1);
		if (p = strchr(ibuf, '\n')
			*p = 0;
		if (!ibuf[0])
			continue;
		for (q = ibuf; *q != ' ' && *q; q++)
			;
		*q++ = '\0';
		while (*q == ' ')
			q++;
		for (pp = p_list; pp; pp = pp->next)
			if (!strcmp(pp->pname, ibuf))
				break;
		if (!pp) {
			printf("unknown pool %s\n", ibuf);
			continue;
		}
		l = dyna_alloc(q, pp);   NTOA(tmp, l);
		printf("alloc for %s: %s\n", ibuf, tmp);
	}
}
#endif
